/*
 * This file is part of the LibreOffice project.
 *
 * Based on LLVM/Clang.
 *
 * This file is distributed under the University of Illinois Open Source
 * License. See LICENSE.TXT for details.
 *
 */

#include "createinstancecheck.hxx"

#include <clang/Basic/SourceManager.h>

using namespace clang;
using namespace std;

#include <iostream>

namespace loplugin
{

CreateInstanceCheck::CreateInstanceCheck( ASTContext& context )
    : context( context )
    {
    }

void CreateInstanceCheck::run()
    {
    TraverseDecl( context.getTranslationUnitDecl());
    }

bool CreateInstanceCheck::VisitCXXMemberCallExpr( CXXMemberCallExpr* expression )
    {
    // TODO also LO header files? or a subdir?
    if( !context.getSourceManager().isFromMainFile( expression->getLocStart()))
        return true;
    // (checks ordered from the cheapest one)
    if( expression->getMethodDecl()->getDeclName().getAsString() != "createInstance" )
        return true;
#if 1 // TODO Figure out a list of the classes to check.
    // TODO Check how to handle inheritance, if actually needed.
    if( expression->getRecordDecl()->getQualifiedNameAsString() != "com::sun::star::lang::XMultiServiceFactory" )
        return true;
#endif
    cout << "X:" << expression->getRecordDecl()->getQualifiedNameAsString() << "::"
        << expression->getMethodDecl()->getName().data() << ":" << expression->getNumArgs() << endl;
    // TODO also check here somehow that the check above actually checks for something
    // that exists (e.g. in case the function gets renamed).
    string argument = getStringFromExpression( expression->getArg( 0 ));
    cout << "ARG:" << argument << endl;
    return true;
    }

string CreateInstanceCheck::getStringFromExpression( const Expr* expression ) const
    {
    // Dive into the expression to find if it's a string literal. There may be
    // various conversions, temporaries, OUString involved and other stuff.
    // casts don't change the value it's created from
    if( const CastExpr* cast = dyn_cast< CastExpr >( expression ))
        return getStringFromExpression( cast->getSubExpr());
    switch( expression->getStmtClass())
        {
        case Stmt::StringLiteralClass:
            {
            const StringLiteral* literal = cast<StringLiteral>( expression );
            if( literal->getKind() == StringLiteral::Ascii )
                return literal->getBytes();
            // TODO: It's a question if this can ever be correct.
            DiagnosticsEngine& diag = context.getDiagnostics();
            // TODO WaE ?
            unsigned diagid = diag.getCustomDiagID( DiagnosticsEngine::Warning,
                "unsupported string literal kind for getStringFromExpression (plugin needs fixing?) [lo-plugin]" );
            diag.Report( expression->getLocStart(), diagid ) << expression->getSourceRange();
            }
        // temporary expression, doesn't change what the value is created from
        case Stmt::MaterializeTemporaryExprClass:
            return getStringFromExpression( cast< MaterializeTemporaryExpr >( expression )->GetTemporaryExpr());
        // bind to temporary, doesn't change what the value is created from
        case Stmt::CXXBindTemporaryExprClass:
            return getStringFromExpression( cast< CXXBindTemporaryExpr >( expression )->getSubExpr());
        case Stmt::ParenExprClass:
            return getStringFromExpression( cast< ParenExpr >( expression )->getSubExpr());
        case Stmt::CXXConstructExprClass:
        case Stmt::CXXTemporaryObjectExprClass: // special case of CXXConstructExprClass
            {
            const CXXConstructExpr* exp = cast< CXXConstructExpr >( expression );
            const CXXConstructorDecl* ctor = exp->getConstructor();
            if( ctor->getParent()->getQualifiedNameAsString() == "rtl::OUString"
                && exp->getNumArgs() >= 1 )
                { // the 2nd argument may be length, type for SFINAE, or similar
                return getStringFromExpression( exp->getArg( 0 ));
                }
            return ""; // TODO
            }
        // a variable or similar, see if it's const and we can figure out the value
        case Stmt::DeclRefExprClass:
            if( const VarDecl* value = dyn_cast< VarDecl >( cast< DeclRefExpr >( expression )->getDecl()))
                {
                // TODO: This may not be strictly speaking right, the value may not be const and may change, but LO
                // code is sometimes written like that. If this is made into more generic code, this may need changing.
                if( const Expr* init = value->getInit())
                    return getStringFromExpression( init );
                }
            // We can't figure out the value. (TODO: Warn?)
            return "";
        case Stmt::UnaryOperatorClass:
            // The RTL_CONSTASCII_USTRINGPARAM macro uses &arg[0] to make sure the passes string
            // is const char* even if it's string literal (const char[]), to avoid triggering
            // the string literal OUString overloads. Handle this special case.
            {
            const UnaryOperator* unary = cast< UnaryOperator >( expression );
            if( unary->getOpcode() == UO_AddrOf )
                if( const ArraySubscriptExpr* array = dyn_cast< ArraySubscriptExpr >( unary->getSubExpr()))
                    if( const IntegerLiteral* integer = dyn_cast< IntegerLiteral >( array->getIdx()))
                        if( integer->getValue() == 0 )
                            return getStringFromExpression( array->getBase());
            // We can't figure it out. (TODO: Warn?)
            return "";
            }
        default:
            DiagnosticsEngine& diag = context.getDiagnostics();
            // TODO WaE ?
            unsigned diagid = diag.getCustomDiagID( DiagnosticsEngine::Warning,
                "cannot analyze for getStringFromExpression (plugin needs fixing) [lo-plugin]" );
            diag.Report( expression->getLocStart(), diagid ) << expression->getSourceRange();
            // TODO make optional?
            expression->dump();
            cerr << "STMT:" << expression->getStmtClassName() << endl;
            return "";
        }
    }

} // namespace
