/*
 * This file is part of the LibreOffice project.
 *
 * Based on LLVM/Clang.
 *
 * This file is distributed under the University of Illinois Open Source
 * License. See LICENSE.TXT for details.
 *
 */

#include "postfixincrementfix.hxx"

#include <clang/Basic/SourceManager.h>
#include <clang/Rewrite/Rewriter.h>

using namespace clang;

namespace loplugin
{

PostfixIncrementFix::PostfixIncrementFix( ASTContext& context, Rewriter& rewriter )
    : context( context ), rewriter( rewriter )
    {
    }

void PostfixIncrementFix::run()
    {
    TraverseDecl( context.getTranslationUnitDecl());
    }

bool PostfixIncrementFix::VisitFunctionDecl( FunctionDecl* declaration )
    {
    // TODO also LO header files? or a subdir?
    if( !context.getSourceManager().isFromMainFile( declaration->getLocStart()))
        return true;
    if( !declaration->doesThisDeclarationHaveABody())
        return true;
    StmtParents parents;
    fixPostfixOperators( declaration->getBody(), parents );
    return true;
    }

void PostfixIncrementFix::fixPostfixOperators( const Stmt* stmt, StmtParents& parents )
    {
    if( const CXXOperatorCallExpr* op = dyn_cast<CXXOperatorCallExpr>( stmt ))
        { // postfix ++ has two arguments (the operand and the hidden extra int)
        if( op->getOperator() == OO_PlusPlus && op->getNumArgs() == 2 ) 
            fixPostfixOperator( op, parents );
        }
    // For primitive types it would be UnaryOperatorExpr, but probably no good reason to change those.
    parents.push_back( stmt );
    for( ConstStmtIterator it = stmt->child_begin();
         it != stmt->child_end();
         ++it )
        {
        if( *it != NULL ) // some children can be apparently NULL
            fixPostfixOperators( *it, parents );
        }
    assert( parents.back() == stmt );
    parents.pop_back();
    }

void PostfixIncrementFix::fixPostfixOperator( const CXXOperatorCallExpr* op, StmtParents& parents )
    {
    if( !canChangePostfixToPrefix( op, parents, parents.size() - 1 ))
        return;
    // TODO: maybe add () around the expression in less obvious cases?
    // Adding spaces around the moved ++ should not be necessary
    // (there might a problem with e.g. a+b++ -> a+++b (i.e. a++ +b),
    // but we don't change such expressions).
    if( rewriter.RemoveText( op->getCallee()->getSourceRange())
        || rewriter.InsertText( op->getLocStart(), "++" ))
        { // the logic is backwards, true here meant it failed, so report
        DiagnosticsEngine& diag = context.getDiagnostics();
        unsigned diagid = diag.getCustomDiagID( DiagnosticsEngine::Warning,
            "cannot fix operator++ (result of macro expansion?) [lo-plugin]" );
        diag.Report( op->getLocStart(), diagid );
        }
    }

bool PostfixIncrementFix::canChangePostfixToPrefix( const CXXOperatorCallExpr* op, StmtParents& parents, int parent_pos )
    {
    // check if foo++ can be safely replaced by ++foo
    switch( parents[ parent_pos ]->getStmtClass())
        {
        case Stmt::CompoundStmtClass:
        case Stmt::ForStmtClass:
            return true;
        // these mean somebody is going to use it
        case Stmt::ImplicitCastExprClass:
        case Stmt::MaterializeTemporaryExprClass:
            return false;
        case Stmt::CXXBindTemporaryExprClass:
            // tricky, it may just mean the temporary will be cleaned up
            // (by ExprWithCleanups), ignore and go up
            assert( parent_pos > 0 ); // should not happen
            return canChangePostfixToPrefix( op, parents, parent_pos - 1 );
        case Stmt::ExprWithCleanupsClass:
            // cleanup of a temporary, should be harmless (if the use
            // of the postfix ++ operator here relies on the fact that
            // the dtor for the object will be called, that's pretty insane
            // code). Ignore and go up.
            assert( parent_pos > 0 ); // should not happen
            return canChangePostfixToPrefix( op, parents, parent_pos - 1 );
        case Stmt::IfStmtClass:
            // cannot be changed in condition, can be changed in statements
            {
            const IfStmt* ifstmt = cast< IfStmt >( parents[ parent_pos ] );
            if( parent_pos == parents.size() - 1 )
                return op != ifstmt->getCond();
            else
                { // indirect child
                assert( parent_pos + 1 < parents.size());
                return parents[ parent_pos + 1 ] != ifstmt->getCond();
                }
            }
        default:
            DiagnosticsEngine& diag = context.getDiagnostics();
            // TODO WaE ?
            unsigned diagid = diag.getCustomDiagID( DiagnosticsEngine::Warning,
                "cannot analyze operator++ (plugin needs fixing) [lo-plugin]" );
            diag.Report( op->getLocStart(), diagid ) << parents[ parent_pos ]->getSourceRange();
            // TODO make optional?
            parents[ std::max( parent_pos - 3, 0 ) ]->dump();
            return false;
        }
    }

} // namespace
